// Copyright (C) 2006-2007 CEREA
//     Author(s): Irne Korsakissok
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/


#ifndef POLYPHEMUS_FILE_DRIVER_PINGDRIVER_HXX


#include "GaussianPuff.cxx"
#include "Puff.cxx"
#include "BaseModel.cxx"
#include <map>
#include <list>
#include <vector>
#include <string>
#include "AtmoData.hxx"

namespace Polyphemus
{


  using namespace std;
  using namespace AtmoData;


  //////////////////////////
  // PLUME-IN-GRID DRIVER //
  //////////////////////////


  /*! \brief This class provides a driver to perform a plume-in-grid
    simulation.
   */
  /*! The driver is responsible for the model initialization, for the loop
    over all meteorological conditions and over all time steps, and for the
    calls to the output saver. It also Its reference floating-point precision
    is 'T'. The model is an instance of 'ClassModel' and the output saver is
    an instance of 'ClassOutputSaver'.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  class PinGDriver: public BaseModel<T>
  {

  protected:

    /*** Main components ***/

    //! Underlying model.
    ClassEulerianModel Model;
    //! Output saver.
    ClassOutputSaver OutputSaver;

    /*** Configuration ***/

    //! Configuration stream.
    ConfigStream config;

    //! Configuration file for gaussian model.
    string gaussian_file;

    //! Display options.
    map<string, bool> option_display;
    
    //! Cartesian or longitude/latitude.
    bool option_cartesian;

    //! Vertical levels management.
    bool altitude_field;

    //! File containing major elevated point sources emissions.
    string source_file;
    //! Number of major elevated point sources emissions.
    int Nsource;
    //! Major point sources emissions indices in Eulerian Grid.
    vector<map<string, int> > source_indices;
    /*! \brief Major point sources emissions coordinates
      (longitude, latitude, height).*/
    vector<map<string, T> > source_coordinates;
    //! List of puff quantities for each source.
    vector<T> quantity;
    vector<string> species;
    //! List of puff models (one for each major point source emission).
    vector<GaussianPuff<T>* > GaussianPuffList;

    //! Eulerian time step, number of time steps and beginning date.
    T delta_t_eul;
    int Nt_eul;
    Date date_min;

    //! Local time step, number of time steps and beginning date.
    T delta_t_local;
    int Nt_local;

    /*** Meteorological data ***/
    
    //! Zonal wind at current date.
    Data<T, 3> ZonalWind_i;
    Data<T, 3> ZonalWind_i_interp;
    //! Meridional wind at current date.
    Data<T, 3> MeridionalWind_i;
    Data<T, 3> MeridionalWind_i_interp;
   //! Temperature at current date.
    Data<T, 3> Temperature_i;

    //! First level wind at current date.
    Data<T, 2> FirstLevelWind_i;
    //! Low cloudiness at current date.
    Data<T, 2> LowCloudiness_i;
    //! Medium cloudiness at current date.
    Data<T, 2> MediumCloudiness_i;
    //! High cloudiness at current date.
    Data<T, 2> HighCloudiness_i;
    //! Insolation at current date.
    Data<T, 2> Insolation_i;
    
    //! Zonal wind buffer.
    Data<T, 3> FileZonalWind_i;
    //! Zonal wind buffer.
    Data<T, 3> FileZonalWind_f;
    //! Meridional wind buffer.
    Data<T, 3> FileMeridionalWind_i;
    //! Meridional wind buffer.
    Data<T, 3> FileMeridionalWind_f;
    //! Temperature buffer.
    Data<T, 3> FileTemperature_i;
    //! Temperature buffer.
    Data<T, 3> FileTemperature_f;

    //! First level wind buffer.
    Data<T, 2> FileFirstLevelWind_i;
    //! First level wind buffer.
    Data<T, 2> FileFirstLevelWind_f;
    //! Low cloudiness buffer.
    Data<T, 2> FileLowCloudiness_i;
    //! Low cloudiness buffer.
    Data<T, 2> FileLowCloudiness_f;
    //! Medium cloudiness buffer.
    Data<T, 2> FileMediumCloudiness_i;
    //! Medium cloudiness buffer.
    Data<T, 2> FileMediumCloudiness_f;
    //! High cloudiness buffer.
    Data<T, 2> FileHighCloudiness_i;
    //! High cloudiness buffer.
    Data<T, 2> FileHighCloudiness_f;
    //! Insolation buffer.
    Data<T, 2> FileInsolation_i;
    //! Insolation buffer.
    Data<T, 2> FileInsolation_f;

  public:

    /*** Constructor and destructor ***/

    PinGDriver(string config_file);
    virtual ~PinGDriver();

    /*** Other methods ***/

    void Allocate();
    virtual void Init();
    virtual void InitStep();
    void ExtractMeteo(map<string, Data<T, 3>* > D3_map,
		      map<string, Data<T, 2>* > D2_map,
		      int index_z, int index_y, int index_x,
		      T lon, T lat, Date date, map<string, T>& met_data,
		      string& stability, bool& isday);
    void ComputeCellWidth(int k, int j, int i, T& CellWidth_z,
			  T& CellWidth_y, T& CellWidth_x, T& CellVolume);
    void GetCellIndices(T lon, T lat, T height,
			int& index_z, int& index_y, int& index_x);
    void TransformZonalWind(Data<T, 3>& ZonalWind);
    void TransformMeridionalWind(Data<T, 3>& MeridionalWind);
    virtual void Run();

  };


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_DRIVER_PINGDRIVER_HXX
#endif
