// Copyright (C) 2005-2007 CEREA
//     Author(s): Irne Korsakissok, Vivien Mallet
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/

// This file is part of a Gaussian puff model for Polyphemus.


#ifndef POLYPHEMUS_FILE_MODELS_GAUSSIANPUFF_HXX


#include "Puff.cxx"
#include <list>
#include "BaseModel.cxx"


namespace Polyphemus
{


  using namespace std;


  //////////////////
  // GAUSSIAN PUFF//
  //////////////////


  //! Gaussian puff model.
  template<class T>
  class GaussianPuff: public BaseModel<T>
  {

  protected:

    //! Current time.
    T time;

    //! Parameterization to compute standard deviations ("Briggs", "Doury").
    string sigma_formula;

    //! Land category ("rural" or "urban") .
    string land_category;

    //! Time ("night" or "day").
    string day_night;

    //! Puff list.
    list<Puff<T> > PuffList;
    //! Number of puffs.
    int Npuff;


    //! List of species half-lives (s).
    Array<T, 1> half_life_time;

    //! List of species biological half-lives (s).
    Array<T, 1> biological_half_life_time;

    //! List of species scavenging coefficients ( s^(-1) ).
    Array<T, 1> scavenging_coefficient;

    //! List of species deposition velocities (m/s).
    Array<T, 1> deposition_velocity;
    /*! \brief Additional number of points per step to compute the Chamberlain
      integral. */
    int Nchamberlain;
    //! Number of points to compute the Chamberlain integral at current step.
    int Nchamberlain_tot;
    //! Deposition model.
    string deposition_model;

    //! Temperature of ambient air (Celsius degrees).
    T temperature_;
    //! Angle between positive x-axis and wind, counterclockwise.
    T wind_angle_;
    //! Wind (m/s).
    T wind_;
    //! Inversion height (m).
    T inversion_height_;

    //! Cosine of wind angle.
    T cos_angle_;
    //! Sine of wind angle.
    T sin_angle_;

    //! Stability class (A, B, C, D, E or F).
    string stability_class_;
    //! Stability (corresponding integer: from 0 for A to 5 for F).
    int stability_;

  public:

    /*** Constructors ***/

    GaussianPuff();
    GaussianPuff(string config_file);

    /*** Other methods ***/

    virtual void Init();
    virtual void InitStep();
    virtual void AddPuffSource(T time_puff, T quantity,T abscissa,
				T ordinate, T height,string species_name);
    const list<Puff<T> >& GetPuffList() const;
    void SetPuffList(list<Puff<T> > pufflist);
    virtual void InitPuffSource();
    virtual void InitPosition();

    virtual void InitMeteo(ConfigStream& meteo, bool show_meteo);
    virtual void SetMeteo(T temperature, T wind_angle, T wind,
			   string stability_class, T inversion_height);
    virtual void SetDayNight(bool isday);

    virtual T GetConcentration(int species, T z, T y, T x);
    using BaseModel<T>::GetConcentration;

    void Advection();
    void Diffusion();
    void Forward();
  };


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODELS_GAUSSIANPUFF_HXX
#endif
