// Copyright (C) 2007 CEREA
//     Author(s): Vivien Mallet
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/


#ifndef POLYPHEMUS_FILE_MODULES_TRANSPORT_SPLITADVECTIONDST3_CXX


#include "SplitAdvectionDST3.hxx"

namespace Polyphemus
{


  //! Initialization of the scheme.
  /*! \note Empty method.
   */
  template<class T>
  template<class ClassModel>
  void SplitAdvectionDST3<T>::Init(ClassModel& Model)
  {
  }


  //! Performs an integration over one time step.
  /*!
    \param Model (input/output) model to be updated. Its interface must
    contain:
    <ul>
    <li> GetDelta_t()
    <li> GetNs()
    <li> GetNz()
    <li> GetNy()
    <li> GetNx()
    <li> HasBoundaryCondition(int)
    <li> A3("BoundaryCondition_z")
    <li> A4("BoundaryCondition_y")
    <li> A4("BoundaryCondition_x")
    <li> D3("ZonalWind")
    <li> D3("MeridionalWind")
    <li> D3("VerticalWind")
    <li> CellWidth_x
    <li> CellWidth_y
    <li> CellWidth_z
    <li> GetConcentration()
    </ul>
  */
  template<class T>
  template<class ClassModel>
  void SplitAdvectionDST3<T>::Forward(ClassModel& Model)
  {
    // Computes the number of subcycles in each direction in order to satisfy
    // the CFL.
    int Nt_x, Nt_y, Nt_z;
    T limit_Delta_t;
    limit_Delta_t = min(Model.CellWidth_x)
      / (max(-Model.D3("ZonalWind").GetMin(),
	    Model.D3("ZonalWind").GetMax()) + 1.e-10);
    Nt_x = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    limit_Delta_t = min(Model.CellWidth_y)
      / (max(-Model.D3("MeridionalWind").GetMin(),
	    Model.D3("MeridionalWind").GetMax()) + 1.e-10);
    Nt_y = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    limit_Delta_t = min(Model.CellWidth_z)
      / (max(-Model.D3("VerticalWind").GetMin(),
	     Model.D3("VerticalWind").GetMax()) + 1.e-10);
    Nt_z = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    
    Data<T, 3> BoundaryCondition_x_i(shape(Model.GetNz(), Model.GetNy(), 2));
    Data<T, 3> BoundaryCondition_y_i(shape(Model.GetNz(), 2, Model.GetNx()));
    Data<T, 2> BoundaryCondition_z_i(shape(Model.GetNy(), Model.GetNx()));

    int with_bc;
    for(int s = 0; s < Model.GetNs(); s++)
      {
	
	/*** Boundary conditions ***/

	if (Model.HasBoundaryCondition(s))
	  {
	    int bc_s = Model.BoundaryConditionIndex(s);
	    BoundaryCondition_x_i.GetArray()
	      = Model.A4("BoundaryCondition_x")(bc_s, Range::all(),
						Range::all(),
						Range::all()).copy();
	    BoundaryCondition_y_i.GetArray()
	      = Model.A4("BoundaryCondition_y")(bc_s, Range::all(),
						Range::all(),
						Range::all()).copy();
	    BoundaryCondition_z_i.GetArray()
	      = Model.A3("BoundaryCondition_z")(bc_s, Range::all(),
						Range::all()).copy();
	    with_bc = 1;
	  }
	else
	  {
	    BoundaryCondition_x_i.SetZero();
	    BoundaryCondition_y_i.SetZero();
	    BoundaryCondition_z_i.SetZero();

	    with_bc = 0;
	  }

	/*** Concentrations ***/

	Data<T, 3> Concentration(&Model.GetConcentration()(s, 0, 0, 0),
				 shape(Model.GetNz(), Model.GetNy(),
				       Model.GetNx()));

	/*** Numerical integration ***/

	Forward(Model.D3("ZonalWind"), Model.D3("MeridionalWind"),
		Model.D3("VerticalWind"), with_bc, Model.CellWidth_x,
		Model.CellWidth_y, Model.CellWidth_z, Model.GetNx(),
		Model.GetNy(), Model.GetNz(), BoundaryCondition_x_i,
		BoundaryCondition_y_i, BoundaryCondition_z_i,
		Model.GetDelta_t(), Nt_x, Nt_y, Nt_y, Concentration);
      }
  }


  //! Performs an integration over one time step for aerosols.
  /*!
    \param Model (input/output) model to be updated. Its interface must
    contain:
    <ul>
    <li> GetDelta_t()
    <li> GetNs_aer()
    <li> GetNz()
    <li> GetNy()
    <li> GetNx()
    <li> HasBoundaryCondition_aer(int, int)
    <li> A3("BoundaryCondition_aer_z")
    <li> A4("BoundaryCondition_aer_y")
    <li> A4("BoundaryCondition_aer_x")
    <li> D3("ZonalWind")
    <li> D3("MeridionalWind")
    <li> D3("VerticalWind")
    <li> CellWidth_x
    <li> CellWidth_y
    <li> CellWidth_z
    <li> GetConcentration_aer()
    </ul>
  */
  template<class T>
  template<class ClassModel>
  void SplitAdvectionDST3<T>::Forward_aer(ClassModel& Model)
  {
    // Computes the number of subcycles in each direction in order to satisfy
    // the CFL.
    int Nt_x, Nt_y, Nt_z;
    T limit_Delta_t;
    limit_Delta_t = min(Model.CellWidth_x)
      / (max(-Model.D3("ZonalWind").GetMin(),
	    Model.D3("ZonalWind").GetMax()) + 1.e-10);
    Nt_x = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    limit_Delta_t = min(Model.CellWidth_y)
      / (max(-Model.D3("MeridionalWind").GetMin(),
	    Model.D3("MeridionalWind").GetMax()) + 1.e-10);
    Nt_y = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    limit_Delta_t = min(Model.CellWidth_z)
      / (max(-Model.D3("VerticalWind").GetMin(),
	     Model.D3("VerticalWind").GetMax()) + 1.e-10);
    Nt_z = int(ceil(Model.GetDelta_t() / limit_Delta_t));
    
    Data<T, 3> BoundaryCondition_x_i(shape(Model.GetNz(), Model.GetNy(), 2));
    Data<T, 3> BoundaryCondition_y_i(shape(Model.GetNz(), 2, Model.GetNx()));
    Data<T, 2> BoundaryCondition_z_i(shape(Model.GetNy(), Model.GetNx()));

    int with_bc;
    vector<int> index;
    int bc_s, bc_b;
    for(int s = 0; s < Model.GetNs_aer(); s++)
      for(int b = 0; b < Model.GetNbin_aer(); b++)
	{
	
	  /*** Boundary conditions ***/

	  if (Model.HasBoundaryCondition_aer(s, b))
	    {
	      index = Model.BoundaryConditionIndex_aer(s, b);
	      bc_s = index[0];
	      bc_b = index[1];
	      BoundaryCondition_x_i.GetArray()
		= Model.A5("BoundaryCondition_x_aer")(bc_s, bc_b,
						      Range::all(),
						      Range::all(),
						      Range::all()).copy();
	      BoundaryCondition_y_i.GetArray()
		= Model.A5("BoundaryCondition_y_aer")(bc_s, bc_b,
						      Range::all(),
						      Range::all(),
						      Range::all()).copy();
	      BoundaryCondition_z_i.GetArray()
		= Model.A4("BoundaryCondition_z_aer")(bc_s, bc_b,
						      Range::all(),
						      Range::all()).copy();
	      with_bc = 1;
	    }
	  else
	    {
	      BoundaryCondition_x_i.SetZero();
	      BoundaryCondition_y_i.SetZero();
	      BoundaryCondition_z_i.SetZero();

	      with_bc = 0;
	    }

	  /*** Concentrations ***/

	  Data<T, 3>
	    Concentration(&Model.GetConcentration_aer()(s, b, 0, 0, 0),
			  shape(Model.GetNz(), Model.GetNy(), Model.GetNx()));

	  /*** Numerical integration ***/

	  Forward(Model.D3("ZonalWind"), Model.D3("MeridionalWind"),
		  Model.D3("VerticalWind"), with_bc, Model.CellWidth_x,
		  Model.CellWidth_y, Model.CellWidth_z, Model.GetNx(),
		  Model.GetNy(), Model.GetNz(), BoundaryCondition_x_i,
		  BoundaryCondition_y_i, BoundaryCondition_z_i,
		  Model.GetDelta_t(), Nt_x, Nt_y, Nt_z, Concentration);
	}
  }


  //! Performs an integration over one time step.
  /*!
    \param ZonalWind zonal wind.
    \param MeridionalWind meridional wind.
    \param VerticalWind vertical wind.
    \param with_bc 1 if boundary conditions are included, 0 otherwise.
    \param CellWidth_x cells widths along x in meters.
    \param CellWidth_y cells widths along y in meters.
    \param CellWidth_z cells widths along z in meters.
    \param Nx number of cells along x.
    \param Ny number of cells along y.
    \param Nz number of cells along z.
    \param BoundaryCondition_x boundary conditions along x.
    \param BoundaryCondition_y boundary conditions along y.
    \param BoundaryCondition_z boundary conditions along z.
    \param Delta_t time step.
    \param Nt_x number of subcycles for the time integration along x.
    \param Nt_y number of subcycles for the time integration along y.
    \param Nt_z number of subcycles for the time integration along z.
    \param Concentration concentrations.
  */
  template<class T>
  void SplitAdvectionDST3<T>::Forward(Data<T, 3>& ZonalWind,
				      Data<T, 3>& MeridionalWind,
				      Data<T, 3>& VerticalWind,
				      int with_bc, Array<T, 1>& CellWidth_x,
				      Array<T, 1>& CellWidth_y,
				      Array<T, 1>& CellWidth_z,
				      int Nx, int Ny, int Nz,
				      Data<T, 3>& BoundaryCondition_x,
				      Data<T, 3>& BoundaryCondition_y,
				      Data<T, 2>& BoundaryCondition_z,
				      T Delta_t, int Nt_x, int Nt_y, int Nt_z,
				      Data<T, 3>& Concentration)
  {
    int k, j, i, h;

    // Subcycle timestep.
    T Delta_t_sub;

    bool bc = with_bc == 1;

    Array<T, 1> Wind(Nx + 1);
    Array<T, 1> BoundaryCondition(2);
    Array<T, 1> Concentration1D(Nx);

    // Along x.
    Delta_t_sub = Delta_t / T(Nt_x);
    for (k = 0; k < Nz; k++)
      for (j = 0; j < Ny; j++)
	{
	  Wind = ZonalWind()(k, j, Range::all());
	  if (bc)
	    {
	      BoundaryCondition(0) = BoundaryCondition_x()(k, j, 0);
	      BoundaryCondition(1) = BoundaryCondition_x()(k, j, 1);
	    }
	  Concentration1D = Concentration()(k, j, Range::all());
	  for (h = 0; h < Nt_x; h++)
	    _split_advection(&Nx, CellWidth_x.data(), Wind.data(),
			     &with_bc, BoundaryCondition.data(),
			     &Delta_t_sub, Concentration1D.data());
	  Concentration()(k, j, Range::all()) = Concentration1D;
	}

    // Along y.
    Delta_t_sub = Delta_t / T(Nt_y);
    Wind.resize(Ny + 1);
    Concentration1D.resize(Ny);
    for (k = 0; k < Nz; k++)
      for (i = 0; i < Nx; i++)
	{
	  Wind = MeridionalWind()(k, Range::all(), i);
	  if (bc)
	    {
	      BoundaryCondition(0) = BoundaryCondition_y()(k, 0, i);
	      BoundaryCondition(1) = BoundaryCondition_y()(k, 1, i);
	    }
	  Concentration1D = Concentration()(k, Range::all(), i);
	  for (h = 0; h < Nt_y; h++)
	    _split_advection(&Ny, CellWidth_y.data(), Wind.data(),
			     &with_bc, BoundaryCondition.data(),
			     &Delta_t_sub, Concentration1D.data());
	  Concentration()(k, Range::all(), i) = Concentration1D;
	}

    // Along z.
    Delta_t_sub = Delta_t / T(Nt_z);
    Wind.resize(Nz + 1);
    Concentration1D.resize(Nz);
    for (j = 0; j < Ny; j++)
      for (i = 0; i < Nx; i++)
	{
	  Wind = VerticalWind()(Range::all(), j, i);
	  if (bc)
	    {
	      BoundaryCondition(0) = BoundaryCondition_z()(0, j, i);
	      BoundaryCondition(1) = BoundaryCondition_z()(1, j, i);
	    }
	  Concentration1D = Concentration()(Range::all(), j, i);
	  for (h = 0; h < Nt_z; h++)
	    _split_advection(&Nz, CellWidth_z.data(), Wind.data(),
			     &with_bc, BoundaryCondition.data(),
			     &Delta_t_sub, Concentration1D.data());
	  Concentration()(Range::all(), j, i) = Concentration1D;
	}
  }


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODULES_TRANSPORT_SPLITADVECTIONDST3_CXX
#endif
