#!/usr/bin/env python

# Copyright (C) 2007 CEREA
#
# CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
# (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
#
# This file is part of the air quality modeling system Polyphemus.
#
# Polyphemus is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# Polyphemus is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# For more information, please see the Polyphemus web site:
#     http://www.enpc.fr/cerea/polyphemus/

#
# This script processes Gocart raw data files into Polair3D input files.
# Gocart concentrations are expected to be daily averages.

import os
import sys
from optparse import OptionParser
import ConfigParser

prog=os.path.basename(sys.argv[0]) + " : "

#################
# CONFIGURATION #

# Fill in the year.
year = 2001
# List all species. Warning: the order is crucial; in case you need 
# to change it, please contact polyphemus@cerea.enpc.fr.
species_list = ["CC.STD.tv15", "DU.STD.tv20", "SU.STD.tv15", "SS.STD.tv17"]

# Put your file format, where %Y is the year, %M is the month
# and %S is the species name.
file_name_format = "/u/cergrene/Y/fahey/CHIN-BCs/%Y%M.%S.g.day"

# Configuration files.
cfg_general_default = "../general.cfg"

# configuration files corresponding to species 
# in "species_list" (same order !).
cfg_species_list = ["bc-gocart-CC.cfg", "bc-gocart-DU.cfg",
                    "bc-gocart-SU.cfg", "bc-gocart-SS.cfg"]

# configuration file for nh4.
cfg_nh4 = "bc-nh4.cfg"

# bc-gocart related program
bc_gocart_prog = "./bc-gocart"
bc_nh4_prog = "./bc-nh4"

# CONFIGURATION #
#################


##########
# SCRIPT #
usage = "%prog [options] general_config_file first_month last_month \n\n" \
	+ " Computes aerosol Polair3D boundary conditions" \
	+ " from Gocart data.\n" \
	+ " general_config_file is optional, default is " + cfg_general_default + " .\n" \
	+ " first_month: index of first month to perform.\n" \
	+ " last_month:  index of last  month to perform.\n" \
	+ " 1=january, ..., 12=december."
version= "%prog 1.0"

parser = OptionParser(usage=usage,version=version)

parser.add_option("-n", "--dry-run", action="count", \
		  dest="dry_run", default=0, help="says what it would do," \
		  + " but actually does NOT do it.")
parser.add_option("--nh4", action="store_true", \
		  dest="nh4", default=False, help="compute ammonia" \
		  + " which is not directly provided by Gocart," \
		  + " default is to not compute.")

(options, args) = parser.parse_args()

# Test number of options
if len(args) != 3 and len(args) != 2:
	os.system(sys.argv[0] + " -h")
	sys.exit(2)

# general config file
if len(args) == 3:
      cfg_general = args[0]
elif len(args) == 2:
	cfg_general = cfg_general_default

if os.path.exists(cfg_general):
	print prog, "using general config file " + cfg_general
else:
	print prog, "general config file " + cfg_general + " NOT found !!"
	sys.exit(2)
	
# Month range.
if len(args) == 3:
	month_range = [int(args[1]), int(args[2]) ]
elif len(args) == 2:
	month_range = [int(args[0]), int(args[1]) ]

print prog, "using month range from " \
      + str(month_range[0]) + " to " + str(month_range[1])

month_range = range(month_range[0], month_range[1] + 1)
file_name_format = file_name_format.replace("%Y", str(year))

# Dry run option.
if options.dry_run > 0:
	print prog, "This is a DRY run, nothing actually performed.\n"

# Process one given year, that of Gocart files.
str_year = str(year)
print prog, "processing year", str_year


# Process mont per month, because of Gocart files.
for month in month_range:
	if month < 10:
		str_month = "0" + str(month)
	else:
		str_month = str(month)
	
	print prog, "month", month
	sys.stdout.flush()
	
        for species, cfg in zip(species_list, cfg_species_list):
			
		file_name = file_name_format.replace("%M", str_month)
		file_name = file_name.replace("%S", species)

		command = bc_gocart_prog + " " + cfg_general + " " \
			  + cfg + " " + file_name + " " \
			  + str_year + str_month

		if options.dry_run > 0:
			print prog + command
		else:
                        print prog, " compute species", species, "..."
		        sys.stdout.flush()
		      	if os.system(command) != 0:
				print " ERROR "
				sys.exit(2)
			else:
				print prog, "... done"

			sys.stdout.flush()


# Perform NH4 for all month range.
if options.nh4:
	command = bc_nh4_prog + " " + cfg_general \
		  + " " + cfg_nh4
	
	if options.dry_run > 0:
		print prog + command
	else:
		print prog, " compute NH4 ..."
		sys.stdout.flush()
		if os.system(command) != 0:
			print " ERROR "
			sys.exit(2)
		else:
			print "done"
			
		sys.stdout.flush()

# SCRIPT #
##########
